#include "som_learn.h"
#include "env.h"

#define MIN(a, b) ((a) <= (b) ? (a) : (b))
#define MAX(a, b) ((a) >= (b) ? (a) : (b))

// ベクトルのマンハッタン距離計算
ufixed16_2 calcManDist(fixed16_2 v1[VECT_DIM], fixed16_2 v2[VECT_DIM]) {
#pragma HLS inline

    // ツリー構造で計算
    // 各ステージのバッファ
    ap_ufixed<17, 3> diff[LIMIT_DIM];
#pragma HLS array_partition variable = diff complete

    ap_ufixed<18, 4> sum1[(LIMIT_DIM >> 1)];
#pragma HLS array_partition variable = sum1 complete

    ap_ufixed<19, 5> sum2[(LIMIT_DIM >> 2)];
#pragma HLS array_partition variable = sum2 complete

    ap_ufixed<20, 6> sum3[(LIMIT_DIM >> 3)];
#pragma HLS array_partition variable = sum3 complete

    ap_ufixed<21, 7> sum4[(LIMIT_DIM >> 4)];
#pragma HLS array_partition variable = sum4 complete

    ap_ufixed<22, 8> sum5[(LIMIT_DIM >> 5)];
#pragma HLS array_partition variable = sum5 complete

    ap_ufixed<23, 9> sum;

    // 各次元の絶対値の差を並列計算
    for (int i = 0; i < VECT_DIM; i++) {
#pragma HLS unroll
        diff[i] = MAX(v1[i], v2[i]) - MIN(v1[i], v2[i]);
    }

    //不要な部分を0埋め
    for (int i = VECT_DIM; i < LIMIT_DIM; i++) {
#pragma HLS unroll
        diff[i] = 0;
    }

    // Stage 1 (64->32)
    for (int j = 0, k = 0; j < (LIMIT_DIM >> 1); j++, k += 2) {
#pragma HLS unroll
        sum1[j] = diff[k] + diff[k + 1];
    }

    // Stage 2 (32->16)
    for (int j = 0, k = 0; j < (LIMIT_DIM >> 2); j++, k += 2) {
#pragma HLS unroll
        sum2[j] = sum1[k] + sum1[k + 1];
    }

    // Stage 3 (16->8)
    for (int j = 0, k = 0; j < (LIMIT_DIM >> 3); j++, k += 2) {
#pragma HLS unroll
        sum3[j] = sum2[k] + sum2[k + 1];
    }

    // Stage 4 (8->4)
    for (int j = 0, k = 0; j < (LIMIT_DIM >> 4); j++, k += 2) {
#pragma HLS unroll
        sum4[j] = sum3[k] + sum3[k + 1];
    }

    // Stage 5 (4->2)
    for (int j = 0, k = 0; j < (LIMIT_DIM >> 5); j++, k += 2) {
#pragma HLS unroll
        sum5[j] = sum4[k] + sum4[k + 1];
    }

    // Stage 6
    sum = sum5[0] + sum5[1];

    return sum;
}

/// <summary>
/// 勝者ニューロンの探索
/// </summary>
/// <param name="vectMap">ベクトルマップ</param>
/// <param name="vect">選択した学習データのベクトル</param>
/// <param name="winnerVect">選択した勝者ニューロンを格納する戻り値</param>
void searchWinner(fixed16_2 vectMap[MAP_SIZE][MAP_SIZE][VECT_DIM], fixed16_2 vect[VECT_DIM], int winnerVect[2]) {
    /// <summary>
    /// 現在計算した勝者との距離を格納
    /// </summary>
    ufixed16_2 currentDist[PARALLEL_LEVEL] = {};
    /// <summary>
    /// 今まで計算した勝者との距離の最小値を格納
    /// </summary>
    ufixed16_2 minDist = 4;
    /// <summary>
    /// 列方向のインデックス
    /// </summary>
    int currentIdxJ[PARALLEL_LEVEL] = {};

    // すべてのベクトルに対してそれぞれ勝者との距離を計算して最小値を探す
    for (int i = 0; i < MAP_SIZE; i++) {
        for (int j = 0; j < NEURON_NUM; j++) {
#pragma HLS pipeline rewind

            // ニューロンに対し、演算を並列化
            for (int k = 0; k < PARALLEL_LEVEL; k++) {
#pragma HLS unroll
                // 列方向のインデックスは（マップサイズの1辺）/（並列度）ごとにインクリメントされる
                currentIdxJ[k] = j + NEURON_NUM * k;
                currentDist[k] = calcManDist(vectMap[i][currentIdxJ[k]], vect);
            }

            // 今までの最小距離より小さかったら最小距離と勝者を更新
            // 上記で並列計算した結果から、さらに勝者を決定する
            for (int loop = PARALLEL_LEVEL >> 1; loop > 1; loop >>= 1) {
                for (int k = 0, l = 0; k < loop; k++, l += 2) {
                    if (currentDist[l] < currentDist[l + 1]) {
                        currentDist[k] = currentDist[l];
                        currentIdxJ[k] = currentIdxJ[l];
                    } else {
                        currentDist[k] = currentDist[l + 1];
                        currentIdxJ[k] = currentIdxJ[l + 1];
                    }
                }
            }

            if (currentDist[0] < minDist) {
                minDist = currentDist[0];
                winnerVect[0] = i;
                winnerVect[1] = currentIdxJ[0];
            }
        }
    }
}

/// <summary>
/// 端と端で繋がっている空間における2点間の距離を求める
/// </summary>
/// <param name="a">座標1</param>
/// <param name="b">座標2</param>
/// <returns>a--b間のトーラス距離</returns>
char calcTorusDistSub(char a, char b) {
#pragma HLS inline
    unsigned char w[4];
#pragma HLS array_partition variable = w complete
    // トーラス距離のパターンは以下の4種類
    w[0] = a - b;
    w[1] = b - a;
    w[2] = a - b + MAP_SIZE;
    w[3] = b - a + MAP_SIZE;

    // 4つの中からトーナメント方式で決定
    for (int i = 0; i < 2; i++) {
#pragma HLS unroll
        w[i] = w[i << 1] <= w[(i << 1) + 1] ? w[(i << 1)] : w[(i << 1) + 1];
    }
    w[0] = w[0] <= w[1] ? w[0] : w[1];
    return w[0];
}

/// <summary>
/// 2点間のトーラス距離を求める
/// </summary>
/// <param name="x1">ベクトル1のX座標</param>
/// <param name="y1">ベクトル1のY座標</param>
/// <param name="x2">ベクトル2のX座標</param>
/// <param name="y2">ベクトル2のY座標</param>
/// <returns>トーラス距離</returns>
inline char calcTorusDist(char x1, char y1, char x2, char y2) {
#pragma HLS pipeline
    return calcTorusDistSub(x1, x2) + calcTorusDistSub(y1, y2);
}

/// <summary>
/// 更新処理のベクトル演算部
/// </summary>
/// <param name="vectMap">ベクトルマップ</param>
/// <param name="selectedVect">学習データ [VECT_DIM]</param>
/// <param name="dist">勝者とのトーラス距離</param>
/// <param name="lut">近傍関数テーブルから参照した値</param>
/// <param name="range">トーラス距離の最大値</param>
void updateVectMapSub(fixed16_2 vectMap[VECT_DIM], fixed16_2 selectedVect[VECT_DIM], uint_16 dist,
                      fixed16_2 lut[(int)(SCOPE + 1)], uint_16 range) {
#pragma HLS inline
    // ベクトルの次元に対し、並列化
    for (int k = 0; k < VECT_DIM; k++) {
#pragma HLS unroll
        fixed16_2 vectDiff = selectedVect[k] - vectMap[k];
        fixed32_4 updateValue = lut[dist] * vectDiff;
        // トーラス距離が一定以上の場合、更新しない
        if (dist <= range) {
            vectMap[k] += (fixed16_2)updateValue;
        }
    }
}

/// <summary>
/// 近傍関数テーブルの更新部
/// </summary>
/// <param name="t">この学習がt回目であることを示す</param>
/// <param name="lut">近傍関数の計算結果を格納するテーブル</param>
/// <param name="range">現在の近傍範囲の最大値</param>
void updateLut(int t, fixed16_2 lut[(int)(SCOPE + 1)], uint_16 &range) {
    /* 係数決定 */
    // 学習が進むにつれて値を減少させる（除算は浮動小数演算）
    fixed32_4 tmp_m = (double)(LEARNING_TIME - t) / LEARNING_TIME;
    // 近傍範囲の計算
    range = (uint_16)SCOPE * (fixed16_2)tmp_m;
    // 学習率を上記で計算したtmp_mにかける
    fixed32_4 m = (fixed16_2)ALPHA * (fixed16_2)tmp_m;

    // パイプラインでテーブルを計算
    // マップ上におけるトーラス距離に応じた値を計算する
    for (int i = 0; i <= SCOPE; i++) {
#pragma HLS pipeline
        // 除算は浮動小数演算
        lut[i] = (fixed16_2)m * (fixed16_2)(1.0 / (1 + i));
    }
}

/// <summary>
/// vectMapの更新
/// </summary>
/// <param name="vectMap">ベクトルマップ</param>
/// <param name="selectedVect">学習データ [VECT_DIM]</param>
/// <param name="winner">勝者ニューロンの1,2次元の添字 {Y,X}</param>
/// <param name="t">この学習がt回目であることを示す</param>
void updateVectMap(fixed16_2 vectMap[MAP_SIZE][MAP_SIZE][VECT_DIM], fixed16_2 selectedVect[VECT_DIM], int winner[2],
                   fixed16_2 lut[(int)(SCOPE + 1)], uint_16 range) {
    /// <summary>
    /// 勝者との距離計算の結果を格納
    /// </summary>
    uint_16 dist[PARALLEL_LEVEL] = {};
#pragma HLS array_partition variable = dist

    for (int i = 0; i < MAP_SIZE; i++) {
        for (int j = 0; j < NEURON_NUM; j++) {
#pragma HLS pipeline

            /* トーラス距離の計算 */
            for (int k = 0; k < PARALLEL_LEVEL; k++) {
#pragma HLS unroll
                dist[k] = calcTorusDist(j + NEURON_NUM * k, i, winner[1], winner[0]);
            }

            /* ベクトルマップの更新 */
            for (int k = 0; k < PARALLEL_LEVEL; k++) {
#pragma HLS unroll
                updateVectMapSub(vectMap[i][j + NEURON_NUM * k], selectedVect, dist[k], lut, range);
            }
        }
    }
}

// DMAでマップと学習データを取得する
#pragma SDS data zero_copy(vectMap [0:MAP_SIZE * MAP_SIZE * VECT_DIM])
#pragma SDS data zero_copy(selectedVect [0:VECT_DIM * HW_TIME])
// 割り当てたメモリが物理的に隣接していることを明示
#pragma SDS data mem_attribute(vectMap : PHYSICAL_CONTIGUOUS)
#pragma SDS data mem_attribute(selectedVect : PHYSICAL_CONTIGUOUS)
void som(fixed16_2 vectMap[MAP_SIZE * MAP_SIZE * VECT_DIM], fixed16_2 selectedVect[VECT_DIM * HW_TIME], const char func) {
    /// <summary> 学習回数 </summary>
    static int t = 0;
    /// <summary> マップのベクトルを格納するBRAM </summary>
    static fixed16_2 map[MAP_SIZE][MAP_SIZE][VECT_DIM];
#pragma HLS array_partition variable = map block factor = 4 dim = 2
#pragma HLS array_partition variable = map complete dim = 3
    /// <summary> 近傍範囲 </summary>
    uint_16 range;
    /// <summary> 近傍関数テーブル </summary>
    fixed16_2 lut[(int)(SCOPE + 1)];
#pragma HLS array_partition variable = lut complete
    switch (func) {
        case SOM_INIT:
            t = 0;
            /* BRAMを初期化 */
            for (int r = 0; r < MAP_SIZE; r++) {
                for (int c = 0; c < MAP_SIZE; c++) {
                    int base = (c + r * MAP_SIZE) * VECT_DIM;
                    for (int i = 0; i < VECT_DIM; i++) {
                        map[r][c][i] = vectMap[base + i];
                    }
                }
            }
            break;
        case SOM_READ:
            /* BRAMのデータを送信 */
            for (int r = 0; r < MAP_SIZE; r++) {
                for (int c = 0; c < MAP_SIZE; c++) {
                    int base = (c + r * MAP_SIZE) * VECT_DIM;
                    for (int i = 0; i < VECT_DIM; i++) {
                        vectMap[base + i] = map[r][c][i];
                    }
                }
            }
            break;
        default:
            for (int i = 0; i < HW_TIME; i++) {
                
                /// <summary> 学習データのベクトル </summary> 
                fixed16_2 vect[VECT_DIM];
                /* DMAで学習データを読む */
#pragma HLS array_partition variable = vect complete
                for (int j = 0; j < VECT_DIM; j++) {
                    vect[j] = selectedVect[i * VECT_DIM + j];
                }

                /* 近傍関数テーブルを更新 */
                updateLut(t, lut, range);

                /// <summary> 勝者ニューロン [Y座標,X座標] </summary>
                int winner[2] = {};

                /* 勝者ニューロンを選択 */
                searchWinner(map, vect, winner);

                /* vectMapの更新 */
                updateVectMap(map, vect, winner, lut, range);

                t++;
            }
            break;
    }
}
